# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import datetime as dt

import pytest
import tenacity

import stamina

from stamina._core import _make_stop


@pytest.mark.parametrize("attempts", [None, 1])
@pytest.mark.parametrize("timeout", [None, 1, dt.timedelta(days=1)])
@pytest.mark.parametrize("duration", [1, dt.timedelta(days=1)])
def test_ok(attempts, timeout, duration):
    """
    No error, no problem.
    """

    @stamina.retry(
        on=Exception,
        attempts=attempts,
        timeout=timeout,
        wait_initial=duration,
        wait_max=duration,
        wait_jitter=duration,
    )
    def f():
        return 42

    assert 42 == f()


@pytest.mark.parametrize(
    "timeout",
    [None, 1, dt.timedelta(days=1)],
)
@pytest.mark.parametrize("duration", [0, dt.timedelta(days=0)])
def test_retries(duration, timeout):
    """
    Retries if the specific error is raised.
    """
    i = 0

    @stamina.retry(
        on=ValueError,
        timeout=timeout,
        wait_max=duration,
        wait_initial=duration,
        wait_jitter=duration,
    )
    def f():
        nonlocal i
        if i < 1:
            i += 1
            raise ValueError

        return 42

    assert 42 == f()
    assert 1 == i


def test_wrong_exception():
    """
    Exceptions that are not passed as `on` are left through.
    """

    @stamina.retry(on=ValueError)
    def f():
        raise TypeError("passed")

    with pytest.raises(TypeError, match="passed"):
        f()


def test_retry_inactive():
    """
    If inactive, don't retry.
    """
    num_called = 0

    @stamina.retry(on=Exception)
    def f():
        nonlocal num_called
        num_called += 1
        raise Exception("passed")

    stamina.set_active(False)

    with pytest.raises(Exception, match="passed"):
        f()

    assert 1 == num_called


def test_retry_inactive_ok():
    """
    If inactive, the happy path still works.
    """
    num_called = 0

    @stamina.retry(on=Exception)
    def f():
        nonlocal num_called
        num_called += 1

    stamina.set_active(False)

    f()

    assert 1 == num_called


def test_retry_block():
    """
    Sync retry_context blocks are retried.
    """
    i = 0

    for attempt in stamina.retry_context(on=ValueError, wait_max=0):
        with attempt:
            i += 1

            assert i == attempt.num
            assert f"<Attempt num={i}>" == repr(attempt)

            if i < 2:
                raise ValueError

    assert 2 == i


class TestMakeStop:
    def test_never(self):
        """
        If all conditions are None, return stop_never.
        """
        assert tenacity.stop_never is _make_stop(attempts=None, timeout=None)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ._data import RetryDetails, RetryHook, RetryHookFactory


def init_structlog() -> RetryHook:
    """
    Initialize structlog instrumentation.

    .. versionadded:: 23.2.0
    """
    import structlog

    logger = structlog.get_logger("stamina")

    def log_retries(details: RetryDetails) -> None:
        logger.warning(
            "stamina.retry_scheduled",
            callable=details.name,
            args=tuple(repr(a) for a in details.args),
            kwargs=dict(details.kwargs.items()),
            retry_num=details.retry_num,
            caused_by=repr(details.caused_by),
            wait_for=round(details.wait_for, 2),
            waited_so_far=round(details.waited_so_far, 2),
        )

    return log_retries


StructlogOnRetryHook = RetryHookFactory(init_structlog)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import importlib.util

import pytest

import stamina._config


@pytest.fixture(autouse=True)
def _reset_config():
    """
    Ensure we're active and have default on-retry hooks before each test.
    """
    stamina.set_active(True)
    stamina.instrumentation.set_on_retry_hooks(None)


BACKENDS = [pytest.param(("asyncio", {}), id="asyncio")]
if importlib.util.find_spec("trio"):
    BACKENDS += [pytest.param(("trio", {}), id="trio")]


@pytest.fixture(params=BACKENDS)
def anyio_backend(request):
    return request.param



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from threading import Lock

from stamina import is_active, set_active
from stamina._config import _Config


def test_activate_deactivate():
    """
    set_active and is_active control/inspect _Active._is_active.
    """
    assert is_active()

    set_active(False)

    assert not is_active()

    set_active(True)

    assert is_active()


def test_config_init_concurrently():
    """
    Config._init_on_retry notices if the hooks have already been initialized
    (presumably while waiting for the lock).
    """
    cfg = _Config(Lock())

    def fake_on_retry(self):
        return self._on_retry

    cfg._get_on_retry = fake_on_retry
    cfg._on_retry = (1, 2)

    assert (1, 2) == cfg._init_on_first_retry()
    assert fake_on_retry is cfg._get_on_retry



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import pytest

import stamina

from stamina.instrumentation import (
    RetryHookFactory,
    get_on_retry_hooks,
    set_on_retry_hooks,
)
from stamina.instrumentation._data import guess_name
from stamina.instrumentation._hooks import get_default_hooks
from stamina.instrumentation._structlog import init_structlog


try:
    import structlog
except ImportError:
    structlog = None

try:
    import prometheus_client
except ImportError:
    prometheus_client = None


def function():
    pass


async def async_function():
    pass


class Foo:
    def method(self):
        pass

    async def async_method(self):
        pass


foo = Foo()


class TestGuessName:
    @pytest.mark.parametrize(
        ("obj", "name"),
        [
            (function, "test_instrumentation.function"),
            (async_function, "test_instrumentation.async_function"),
            (foo.method, "test_instrumentation.Foo.method"),
            (foo.async_method, "test_instrumentation.Foo.async_method"),
            (Foo.method, "test_instrumentation.Foo.method"),
            (Foo.async_method, "test_instrumentation.Foo.async_method"),
        ],
    )
    def test_module_scope(self, obj, name):
        """
        Names of callables are guessed.
        """
        assert f"tests.{name}" == guess_name(obj)

    def test_local(self):
        """
        Scope-local functions are guessed.
        """

        def f():
            pass

        async def async_f():
            pass

        assert (
            "tests.test_instrumentation.TestGuessName.test_local.<locals>.f"
            == guess_name(f)
        )
        assert (
            "tests.test_instrumentation.TestGuessName.test_local.<locals>.async_f"
            == guess_name(async_f)
        )


def test_get_default_hooks():
    """
    Both default instrumentations are detected.
    """
    if prometheus_client:
        assert 2 == len(get_default_hooks())
    else:
        assert 1 == len(get_default_hooks())


def test_get_prometheus_counter():
    """
    Returns finalized counter if active.
    """
    counter = stamina.instrumentation.get_prometheus_counter()

    if prometheus_client:
        assert counter is not None
    else:
        assert counter is None


@pytest.mark.skipif(not structlog, reason="needs structlog")
def test_structlog_detected():
    """
    If structlog is importable, init_structlog returns a callable.
    """
    assert init_structlog()


@pytest.mark.skipif(structlog, reason="needs missing structlog")
class TestLogging:
    def test_sync(self, caplog):
        """
        Sync retries are logged.
        """

        @stamina.retry(on=ValueError, wait_max=0, attempts=2)
        def f():
            raise ValueError

        with pytest.raises(ValueError):
            f()

        assert [
            ("stamina", 30, "stamina.retry_scheduled")
        ] == caplog.record_tuples

    @pytest.mark.anyio()
    async def test_async(self, caplog):
        """
        Async retries are logged.
        """

        @stamina.retry(on=ValueError, wait_max=0, attempts=2)
        async def f():
            raise ValueError

        with pytest.raises(ValueError):
            await f()

        assert [
            ("stamina", 30, "stamina.retry_scheduled")
        ] == caplog.record_tuples


class TestSetOnRetryHooks:
    def test_none_is_default(self):
        """
        None is replaced with default hooks.
        """
        assert get_on_retry_hooks() is not None
        assert () != get_on_retry_hooks()

        set_on_retry_hooks(())

        assert () == get_on_retry_hooks()

        set_on_retry_hooks(None)

        assert () != get_on_retry_hooks()
        assert get_on_retry_hooks() is not None

    def test_init_hooks(self):
        """
        If a hook is wrapped in RetryHookFactory, init_hooks transforms it into
        a RetryHook. Otherwise it's left alone.
        """

        def hook(details):
            pass

        def delayed_hook(details):
            pass

        def init():
            return delayed_hook

        set_on_retry_hooks([hook, RetryHookFactory(init)])

        assert (
            hook,
            delayed_hook,
        ) == get_on_retry_hooks()



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from . import instrumentation
from ._config import is_active, set_active
from ._core import Attempt, retry, retry_context


__all__ = [
    "Attempt",
    "retry",
    "retry_context",
    "is_active",
    "set_active",
    "instrumentation",
]


def __getattr__(name: str) -> str:
    if name != "__version__":
        msg = f"module {__name__} has no attribute {name}"
        raise AttributeError(msg)

    from importlib.metadata import version

    return version("stamina")



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

"""
This file is used to test the type annotations of the public API. It is NOT
meant to be executed.
"""

from __future__ import annotations

import datetime as dt

from stamina import (
    is_active,
    retry,
    retry_context,
    set_active,
)
from stamina.instrumentation import (
    RetryDetails,
    RetryHook,
    RetryHookFactory,
    get_on_retry_hooks,
    set_on_retry_hooks,
)


@retry(on=ValueError)
def just_exc() -> None:
    ...


@retry(on=TypeError)
async def just_exc_async() -> None:
    ...


@retry(on=TypeError, timeout=13.0)
def exc_timeout() -> None:
    ...


@retry(on=TypeError, timeout=dt.timedelta(seconds=13.0))
def exc_timeout_timedelta() -> None:
    ...


@retry(on=TypeError, timeout=13.0, attempts=10)
def exc_timeout_attempts() -> None:
    ...


@retry(on=TypeError, timeout=None, attempts=None)
def exc_timeout_attempts_none() -> None:
    ...


@retry(
    on=TypeError,
    wait_initial=0.3,
    wait_max=1.0,
    wait_jitter=1.0,
    wait_exp_base=2.0,
)
def exc_tune_waiting() -> None:
    ...


@retry(
    on=TypeError,
    wait_initial=1,
    wait_max=2,
    wait_jitter=3,
    wait_exp_base=4,
)
def exc_tune_waiting_ints() -> None:
    ...


one_sec = dt.timedelta(seconds=1.0)


@retry(
    on=TypeError,
    timeout=one_sec,
    wait_initial=one_sec,
    wait_max=one_sec,
    wait_jitter=one_sec,
)
def exc_tune_waiting_timedelta() -> None:
    ...


set_active(False)

if is_active() is True:
    ...

set_active(False)


def hook(details: RetryDetails) -> None:
    return None


def init() -> RetryHook:
    return hook


set_on_retry_hooks([hook, RetryHookFactory(init)])

hooks: tuple[RetryHook, ...] = get_on_retry_hooks()


for attempt in retry_context(on=ValueError, timeout=13):
    with attempt:
        x: int = attempt.num

for attempt in retry_context(
    on=ValueError, timeout=dt.timedelta(seconds=13.0)
):
    with attempt:
        ...


async def f() -> None:
    async for attempt in retry_context(
        on=ValueError,
        timeout=one_sec,
        wait_initial=one_sec,
        wait_max=one_sec,
        wait_jitter=one_sec,
    ):
        with attempt:
            pass



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING

from ._data import RetryDetails, RetryHook, RetryHookFactory, guess_name


if TYPE_CHECKING:
    from prometheus_client import Counter


RETRIES_TOTAL = None


def init_prometheus() -> RetryHook:
    """
    Initialize Prometheus instrumentation.
    """
    from prometheus_client import Counter

    global RETRIES_TOTAL  # noqa: PLW0603

    # Mostly for testing so we can call init_prometheus more than once.
    if RETRIES_TOTAL is None:
        RETRIES_TOTAL = Counter(
            "stamina_retries_total",
            "Total number of retries.",
            ("callable", "retry_num", "error_type"),
        )

    def count_retries(details: RetryDetails) -> None:
        """
        Count and log retries for callable *name*.
        """
        RETRIES_TOTAL.labels(
            callable=details.name,
            retry_num=details.retry_num,
            error_type=guess_name(details.caused_by.__class__),
        ).inc()

    return count_retries


def get_prometheus_counter() -> Counter | None:
    """
    Return the Prometheus counter for the number of retries.

    Returns:
        If active, the Prometheus `counter
        <https://github.com/prometheus/client_python>`_ for the number of
        retries. None otherwise.

    .. versionadded:: 23.2.0
    """
    from . import get_on_retry_hooks

    # Finalize the hooks if not done yet.
    get_on_retry_hooks()

    return RETRIES_TOTAL


PrometheusOnRetryHook = RetryHookFactory(init_prometheus)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, Protocol


def guess_name(obj: object) -> str:
    name = getattr(obj, "__qualname__", None) or "<unnamed object>"
    mod = getattr(obj, "__module__", None) or "<unknown module>"

    if mod == "builtins":
        return name

    return f"{mod}.{name}"


@dataclass(frozen=True)
class RetryDetails:
    r"""
    Details about a retry attempt that are passed into :class:`RetryHook`\ s.

    Attributes:
        name: Name of the callable that is being retried.

        args: Positional arguments that were passed to the callable.

        kwargs: Keyword arguments that were passed to the callable.

        retry_num: Number of the retry attempt. Starts at 1 after the first
            failure.

        wait_for: Time in seconds that *stamina* will wait before the next
            attempt.

        waited_so_far: Time in seconds that *stamina* has waited so far for the
            current callable.

        caused_by: Exception that caused the retry attempt.

    .. versionadded:: 23.2.0
    """

    __slots__ = (
        "name",
        "args",
        "kwargs",
        "retry_num",
        "wait_for",
        "waited_so_far",
        "caused_by",
    )

    name: str
    args: tuple[object, ...]
    kwargs: dict[str, object]
    retry_num: int
    wait_for: float
    waited_so_far: float
    caused_by: Exception


class RetryHook(Protocol):
    """
    A callable that gets called after an attempt has failed and a retry has
    been scheduled.

    This is a :class:`typing.Protocol` that can be implemented by any callable
    that takes one argument of type :class:`RetryDetails` and returns None.

    .. versionadded:: 23.2.0
    """

    def __call__(self, details: RetryDetails) -> None:
        ...


@dataclass(frozen=True)
class RetryHookFactory:
    """
    Wraps a callable that returns a :class:`RetryHook`.

    They are called on the first scheduled retry and can be used to delay
    initialization. If you need to pass arguments, you can do that using
    :func:`functools.partial`.

    .. versionadded:: 23.2.0
    """

    hook_factory: Callable[[], RetryHook]



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from threading import Lock
from typing import Callable

from .instrumentation import RetryHookFactory
from .instrumentation._hooks import get_default_hooks, init_hooks
from .typing import RetryHook


class _Config:
    """
    Global stamina configuration.

    Strictly private.
    """

    __slots__ = ("lock", "_is_active", "_on_retry", "_get_on_retry")

    lock: Lock
    _is_active: bool
    _on_retry: tuple[RetryHook, ...] | tuple[
        RetryHook | RetryHookFactory, ...
    ] | None
    _get_on_retry: Callable[[], tuple[RetryHook, ...]]

    def __init__(self, lock: Lock) -> None:
        self.lock = lock
        self._is_active = True

        # Prepare delayed initialization.
        self._on_retry = None
        self._get_on_retry = self._init_on_first_retry

    @property
    def is_active(self) -> bool:
        return self._is_active

    @is_active.setter
    def is_active(self, value: bool) -> None:
        with self.lock:
            self._is_active = value

    @property
    def on_retry(self) -> tuple[RetryHook, ...]:
        return self._get_on_retry()

    @on_retry.setter
    def on_retry(
        self, value: tuple[RetryHook | RetryHookFactory, ...] | None
    ) -> None:
        with self.lock:
            self._get_on_retry = self._init_on_first_retry
            self._on_retry = value

    def _init_on_first_retry(self) -> tuple[RetryHook, ...]:
        """
        Perform delayed initialization of on_retry hooks.
        """
        with self.lock:
            # Ensure hooks didn't init while waiting for the lock.
            if self._get_on_retry == self._init_on_first_retry:
                if self._on_retry is None:
                    self._on_retry = get_default_hooks()

                self._on_retry = init_hooks(self._on_retry)

                self._get_on_retry = lambda: self._on_retry  # type: ignore[assignment, return-value]

        return self._on_retry  # type: ignore[return-value]


CONFIG = _Config(Lock())


def is_active() -> bool:
    """
    Check whether retrying is active.

    Returns:
        Whether retrying is active.
    """
    return CONFIG.is_active


def set_active(active: bool) -> None:
    """
    Activate or deactivate retrying.

    Is idempotent and can be called repeatedly with the same value.
    """
    CONFIG.is_active = bool(active)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ._data import RetryDetails, RetryHook, RetryHookFactory


def init_logging(log_level: int = 30) -> RetryHook:
    """
    Initialize logging using the standard library.

    Returned hook logs scheduled retries at *log_level*.

    .. versionadded:: 23.2.0
    """
    import logging

    logger = logging.getLogger("stamina")

    def log_retries(details: RetryDetails) -> None:
        logger.log(
            log_level,
            "stamina.retry_scheduled",
            extra={
                "stamina.callable": details.name,
                "stamina.args": tuple(repr(a) for a in details.args),
                "stamina.kwargs": dict(details.kwargs.items()),
                "stamina.retry_num": details.retry_num,
                "stamina.caused_by": repr(details.caused_by),
                "stamina.wait_for": round(details.wait_for, 2),
                "stamina.waited_so_far": round(details.waited_so_far, 2),
            },
        )

    return log_retries


LoggingOnRetryHook = RetryHookFactory(init_logging)



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from ._data import RetryDetails, RetryHook, RetryHookFactory
from ._hooks import get_on_retry_hooks, set_on_retry_hooks
from ._logging import LoggingOnRetryHook
from ._prometheus import PrometheusOnRetryHook, get_prometheus_counter
from ._structlog import StructlogOnRetryHook


__all__ = [
    "get_on_retry_hooks",
    "set_on_retry_hooks",
    "get_prometheus_counter",
    "LoggingOnRetryHook",
    "PrometheusOnRetryHook",
    "RetryDetails",
    "RetryHook",
    "RetryHookFactory",
    "StructlogOnRetryHook",
]



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from importlib import metadata

import stamina


def test_version(recwarn):
    """
    stamina.__version__ returns the correct version and doesn't warn.
    """
    assert metadata.version("stamina") == stamina.__version__
    assert [] == recwarn.list



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Iterable

from ._data import RetryHook, RetryHookFactory
from ._logging import LoggingOnRetryHook
from ._prometheus import PrometheusOnRetryHook
from ._structlog import StructlogOnRetryHook


def init_hooks(
    maybe_delayed: tuple[RetryHook | RetryHookFactory, ...]
) -> tuple[RetryHook, ...]:
    """
    Execute delayed hook factories and return a tuple of finalized hooks.
    """
    hooks = []
    for hook_or_factory in maybe_delayed:
        if isinstance(hook_or_factory, RetryHookFactory):
            hooks.append(hook_or_factory.hook_factory())
        else:
            hooks.append(hook_or_factory)

    return tuple(hooks)


def get_default_hooks() -> tuple[RetryHookFactory, ...]:
    """
    Return the default hooks according to availability.
    """
    hooks = []

    try:
        import prometheus_client  # noqa: F401

        hooks.append(PrometheusOnRetryHook)
    except ImportError:
        pass

    try:
        import structlog  # noqa: F401

        hooks.append(StructlogOnRetryHook)
    except ImportError:
        hooks.append(LoggingOnRetryHook)

    return tuple(hooks)


def set_on_retry_hooks(
    hooks: Iterable[RetryHook | RetryHookFactory] | None,
) -> None:
    """
    Set hooks that are called after a retry has been scheduled.

    Args:
        hooks: Hooks to call after a retry has been scheduled. Passing
            None resets to default. To deactivate instrumentation, pass an
            empty iterable.

    .. versionadded:: 23.2.0
    """
    from .._config import CONFIG

    CONFIG.on_retry = tuple(hooks) if hooks is not None else hooks  # type: ignore[assignment,arg-type]


def get_on_retry_hooks() -> tuple[RetryHook, ...]:
    """
    Get hooks that are called after a retry has been scheduled.

    Returns:
        Hooks that will run if a retry is scheduled. Factories are called if
        they haven't already.

    .. versionadded:: 23.2.0
    """
    from .._config import CONFIG

    return CONFIG.on_retry



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import os
import pathlib
import shutil
import sys

import nox


if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib


nox.options.sessions = ["pre_commit", "tests", "mypy_api", "mypy_pkg"]
nox.options.reuse_existing_virtualenvs = True
nox.options.error_on_external_run = True


pyp = tomllib.loads(pathlib.Path("pyproject.toml").read_text())
ALL_SUPPORTED = [
    pv.rsplit(" ")[-1]
    for pv in pyp["project"]["classifiers"]
    if pv.startswith("Programming Language :: Python :: ")
]


@nox.session
def pre_commit(session: nox.Session) -> None:
    session.install("pre-commit")

    session.run("pre-commit", "run", "--all-files")


@nox.session(python=ALL_SUPPORTED)
def mypy_api(session: nox.Session) -> None:
    session.install(".[typing]", "structlog", "prometheus-client")

    session.run("mypy", "tests/typing")


@nox.session
def mypy_pkg(session: nox.Session) -> None:
    session.install(".[typing]", "structlog", "prometheus-client")

    session.run("mypy", "src", "tests/typing", "noxfile.py")


def _get_pkg(posargs: list[str]) -> tuple[str, list[str]]:
    """
    Allow `--installpkg path/to/wheel.whl` to be passed.
    """
    posargs = list(posargs)

    try:
        i = posargs.index("--installpkg")
        pkg = posargs[i + 1]
        del posargs[i : i + 2]
    except ValueError:
        pkg = "."

    return pkg + "[tests]", posargs


@nox.session(python=ALL_SUPPORTED)
@nox.parametrize(
    "opt_deps", [[], ["structlog", "prometheus-client"], ["trio"]]
)
def tests(session: nox.Session, opt_deps: list[str]) -> None:
    pkg, posargs = _get_pkg(session.posargs)

    session.install(pkg, "coverage[toml]", *opt_deps)

    session.run("coverage", "run", "-m", "pytest", *posargs)

    if os.environ.get("CI") != "true":
        session.notify("coverage_report")


@nox.session
def coverage_report(session: nox.Session) -> None:
    session.install("coverage[toml]")

    session.run("coverage", "combine")
    session.run("coverage", "report")


@nox.session
def docs(session: nox.Session) -> None:
    shutil.rmtree("docs/_build", ignore_errors=True)

    if session.posargs and session.posargs[0] == "watch":
        session.install("-e", ".[docs]", "watchfiles")
        session.run(
            "watchfiles",
            "--ignore-paths",
            "docs/_build",
            "python -Im sphinx "
            "-T -E "
            "-W --keep-going "
            "-b html "
            "-d docs/_build/doctrees "
            "-D language=en "
            "-n "
            "docs "
            "docs/_build/html",
        )
        return

    session.install(".[docs]")
    cmds = session.posargs or ["html", "doctest"]

    for cmd in cmds:
        session.run(
            # fmt: off
            "python", "-Im", "sphinx",
            "-T", "-E",
            "-W", "--keep-going",
            "-b", cmd,
            "-d", "docs/_build/doctrees",
            "-D", "language=en",
            "-n",
            "docs",
            "docs/_build/html",
            # fmt: on
        )



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime as dt
import sys

from collections.abc import Callable
from dataclasses import dataclass, replace
from functools import wraps
from inspect import iscoroutinefunction
from types import TracebackType
from typing import AsyncIterator, Iterator, TypeVar

import tenacity as _t

from ._config import CONFIG, _Config
from .instrumentation._data import RetryDetails, guess_name


if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

try:
    from sniffio import current_async_library
except ImportError:  # pragma: no cover -- we always have sniffio in tests

    def current_async_library() -> str:
        return "asyncio"


async def _smart_sleep(delay: float) -> None:
    io = current_async_library()

    if io == "asyncio":
        import asyncio

        await asyncio.sleep(delay)
    elif io == "trio":
        import trio

        await trio.sleep(delay)
    else:  # pragma: no cover
        msg = f"Unknown async library: {io!r}"
        raise RuntimeError(msg)


T = TypeVar("T")
P = ParamSpec("P")


def retry_context(
    on: type[Exception] | tuple[type[Exception], ...],
    attempts: int | None = 10,
    timeout: float | dt.timedelta | None = 45.0,
    wait_initial: float | dt.timedelta = 0.1,
    wait_max: float | dt.timedelta = 5.0,
    wait_jitter: float | dt.timedelta = 1.0,
    wait_exp_base: float = 2.0,
) -> _RetryContextIterator:
    """
    Iterator that yields context managers that can be used to retry code
    blocks.

    Arguments have the same meaning as for :func:`stamina.retry`.

    .. versionadded:: 23.1.0
    .. versionadded:: 23.3.0 `Trio <https://trio.readthedocs.io/>`_ support.
    """

    return _RetryContextIterator.from_params(
        on=on,
        attempts=attempts,
        timeout=timeout,
        wait_initial=wait_initial,
        wait_max=wait_max,
        wait_jitter=wait_jitter,
        wait_exp_base=wait_exp_base,
        name="<context block>",
        args=(),
        kw={},
    )


class Attempt:
    """
    A context manager that can be used to retry code blocks.

    Instances are yielded by the :func:`stamina.retry_context` iterator.

    .. versionadded:: 23.2.0
    """

    __slots__ = ("_t_attempt",)

    _t_attempt: _t.AttemptManager

    def __init__(self, attempt: _t.AttemptManager):
        self._t_attempt = attempt

    def __repr__(self) -> str:
        return f"<Attempt num={self.num}>"

    @property
    def num(self) -> int:
        """
        The number of the current attempt.
        """
        return self._t_attempt.retry_state.attempt_number  # type: ignore[no-any-return]

    def __enter__(self) -> None:
        return self._t_attempt.__enter__()  # type: ignore[no-any-return]

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> bool | None:
        return self._t_attempt.__exit__(  # type: ignore[no-any-return]
            exc_type, exc_value, traceback
        )


_STOP_NO_RETRY = _t.stop_after_attempt(1)


class _LazyNoAsyncRetry:
    """
    Allows us a free null object pattern using non-retries and avoid None.
    """

    __slots__ = ()

    def __aiter__(self) -> _t.AsyncRetrying:
        return _t.AsyncRetrying(
            reraise=True, stop=_STOP_NO_RETRY, sleep=_smart_sleep
        ).__aiter__()


_LAZY_NO_ASYNC_RETRY = _LazyNoAsyncRetry()


@dataclass
class _RetryContextIterator:
    __slots__ = ("_t_kw", "_t_a_retrying", "_name", "_args", "_kw")
    _t_kw: dict[str, object]
    _t_a_retrying: _t.AsyncRetrying
    _name: str
    _args: tuple[object, ...]
    _kw: dict[str, object]

    @classmethod
    def from_params(
        cls,
        on: type[Exception] | tuple[type[Exception], ...],
        attempts: int | None,
        timeout: float | dt.timedelta | None,
        wait_initial: float | dt.timedelta,
        wait_max: float | dt.timedelta,
        wait_jitter: float | dt.timedelta,
        wait_exp_base: float,
        name: str,
        args: tuple[object, ...],
        kw: dict[str, object],
    ) -> _RetryContextIterator:
        return cls(
            _name=name,
            _args=args,
            _kw=kw,
            _t_kw={
                "retry": _t.retry_if_exception_type(on),
                "wait": _t.wait_exponential_jitter(
                    initial=wait_initial.total_seconds()
                    if isinstance(wait_initial, dt.timedelta)
                    else wait_initial,
                    max=wait_max.total_seconds()
                    if isinstance(wait_max, dt.timedelta)
                    else wait_max,
                    exp_base=wait_exp_base,
                    jitter=wait_jitter.total_seconds()
                    if isinstance(wait_jitter, dt.timedelta)
                    else wait_jitter,
                ),
                "stop": _make_stop(
                    attempts=attempts,
                    timeout=timeout.total_seconds()
                    if isinstance(timeout, dt.timedelta)
                    else timeout,
                ),
                "reraise": True,
            },
            _t_a_retrying=_LAZY_NO_ASYNC_RETRY,
        )

    def with_name(
        self, name: str, args: tuple[object, ...], kw: dict[str, object]
    ) -> _RetryContextIterator:
        """
        Recreate ourselves with a new name and arguments.
        """
        return replace(self, _name=name, _args=args, _kw=kw)

    def __iter__(self) -> Iterator[Attempt]:
        if not CONFIG.is_active:
            for r in _t.Retrying(
                reraise=True, stop=_STOP_NO_RETRY
            ):  # pragma: no cover -- it's always once + GeneratorExit
                yield Attempt(r)

        for r in _t.Retrying(
            before_sleep=_make_before_sleep(
                self._name, CONFIG, self._args, self._kw
            ),
            **self._t_kw,
        ):
            yield Attempt(r)

    def __aiter__(self) -> AsyncIterator[Attempt]:
        if CONFIG.is_active:
            self._t_a_retrying = _t.AsyncRetrying(
                sleep=_smart_sleep,
                before_sleep=_make_before_sleep(
                    self._name, CONFIG, self._args, self._kw
                ),
                **self._t_kw,
            )

        self._t_a_retrying = self._t_a_retrying.__aiter__()

        return self

    async def __anext__(self) -> Attempt:
        return Attempt(await self._t_a_retrying.__anext__())


def _make_before_sleep(
    name: str,
    config: _Config,
    args: tuple[object, ...],
    kw: dict[str, object],
) -> Callable[[_t.RetryCallState], None]:
    """
    Create a `before_sleep` callback function that runs our `RetryHook`s with
    the necessary arguments.
    """

    last_idle_for = 0.0

    def before_sleep(rcs: _t.RetryCallState) -> None:
        nonlocal last_idle_for

        wait_for = rcs.idle_for - last_idle_for

        details = RetryDetails(
            name=name,
            retry_num=rcs.attempt_number,
            wait_for=wait_for,
            waited_so_far=rcs.idle_for - wait_for,
            caused_by=rcs.outcome.exception(),
            args=args,
            kwargs=kw,
        )

        for hook in config.on_retry:
            hook(details)

        last_idle_for = rcs.idle_for

    return before_sleep


def _make_stop(*, attempts: int | None, timeout: float | None) -> _t.stop_base:
    """
    Combine *attempts* and *timeout* into one stop condition.
    """
    stops = []

    if attempts:
        stops.append(_t.stop_after_attempt(attempts))

    if timeout:
        stops.append(_t.stop_after_delay(timeout))

    if len(stops) > 1:
        return _t.stop_any(*stops)

    if not stops:
        return _t.stop_never

    return stops[0]


def retry(
    *,
    on: type[Exception] | tuple[type[Exception], ...],
    attempts: int | None = 10,
    timeout: float | dt.timedelta | None = 45.0,
    wait_initial: float | dt.timedelta = 0.1,
    wait_max: float | dt.timedelta = 5.0,
    wait_jitter: float | dt.timedelta = 1.0,
    wait_exp_base: float = 2.0,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    r"""
    Retry if one of configured exceptions are raised.

    The backoff delays between retries grow exponentially plus a random jitter.

    The backoff for retry attempt number *attempt* is computed as:

    .. keep in-sync with docs/motivation.md
    .. math::

       min(wait\_max, wait\_initial * wait\_exp\_base^{attempt - 1} + random(0, wait\_jitter))

    Since :math:`x^0` is always 1, the first backoff is within the interval
    :math:`[wait\_initial,wait\_initial+wait\_jitter]`. Thus, with default
    values between 0.1 and 1.1 seconds.

    If all retries fail, the *last* exception is let through.

    All float-based time parameters are in seconds.

    Parameters:
        on: An Exception or a tuple of Exceptions on which the decorated
            callable will be retried. There is no default -- you *must* pass
            this explicitly.

        attempts: Maximum total number of attempts. Can be combined with
            *timeout*.

        timeout: Maximum total time for all retries. Can be combined with
            *attempts*.

        wait_initial: Minimum backoff before the *first* retry.

        wait_max: Maximum backoff time between retries at any time.

        wait_jitter: Maximum *jitter* that is added to retry back-off delays
            (the actual jitter added is a random number between 0 and
            *wait_jitter*)

        wait_exp_base: The exponential base used to compute the retry backoff.

    .. versionchanged:: 23.1.0

       All time-related parameters can now be specified as a
       :class:`datetime.timedelta`.

    .. versionadded:: 23.3.0 `Trio <https://trio.readthedocs.io/>`_ support.
    """
    retry_ctx = _RetryContextIterator.from_params(
        on=on,
        attempts=attempts,
        timeout=timeout,
        wait_initial=wait_initial,
        wait_max=wait_max,
        wait_jitter=wait_jitter,
        wait_exp_base=wait_exp_base,
        name="<unknown>",
        args=(),
        kw={},
    )

    def retry_decorator(wrapped: Callable[P, T]) -> Callable[P, T]:
        name = guess_name(wrapped)

        if not iscoroutinefunction(wrapped):

            @wraps(wrapped)
            def sync_inner(*args: P.args, **kw: P.kwargs) -> T:  # type: ignore[return]
                for attempt in retry_ctx.with_name(  # noqa: RET503
                    name, args, kw
                ):
                    with attempt:
                        return wrapped(*args, **kw)

            return sync_inner

        @wraps(wrapped)
        async def async_inner(*args: P.args, **kw: P.kwargs) -> T:  # type: ignore[return]
            async for attempt in retry_ctx.with_name(  # noqa: RET503
                name, args, kw
            ):
                with attempt:
                    return await wrapped(*args, **kw)  # type: ignore[no-any-return]

        return async_inner  # type: ignore[return-value]

    return retry_decorator



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import datetime as dt

import pytest

import stamina


pytestmark = pytest.mark.anyio


@pytest.mark.parametrize("attempts", [None, 1])
@pytest.mark.parametrize("timeout", [None, 1, dt.timedelta(days=1)])
@pytest.mark.parametrize("duration", [1, dt.timedelta(days=1)])
async def test_ok(attempts, timeout, duration):
    """
    No error, no problem.
    """

    class C:
        @stamina.retry(
            on=Exception,
            attempts=attempts,
            timeout=timeout,
            wait_initial=duration,
            wait_max=duration,
            wait_jitter=duration,
        )
        async def f(self):
            return 42

    @stamina.retry(
        on=Exception,
        attempts=attempts,
        timeout=timeout,
        wait_initial=duration,
        wait_max=duration,
        wait_jitter=duration,
    )
    async def f():
        return 42

    assert 42 == await f()
    assert 42 == await C().f()


@pytest.mark.parametrize("timeout", [None, 1, dt.timedelta(days=1)])
@pytest.mark.parametrize("duration", [0, dt.timedelta(days=0)])
async def test_retries(duration, timeout):
    """
    Retries if the specific error is raised.
    """
    i = 0

    @stamina.retry(
        on=ValueError,
        timeout=timeout,
        wait_max=duration,
        wait_initial=duration,
        wait_jitter=duration,
    )
    async def f():
        nonlocal i
        if i == 0:
            i += 1
            raise ValueError

        return 42

    assert 42 == await f()
    assert 1 == i


async def test_retries_method():
    """
    Retries if the specific error is raised.
    """
    i = 0

    class C:
        @stamina.retry(on=ValueError, wait_max=0)
        async def f(self):
            nonlocal i
            if i == 0:
                i += 1
                raise ValueError

            return 42

    assert 42 == await C().f()
    assert 1 == i


async def test_wrong_exception():
    """
    Exceptions that are not passed as `on` are left through.
    """

    @stamina.retry(on=ValueError)
    async def f():
        raise TypeError("passed")

    with pytest.raises(TypeError, match="passed"):
        await f()


async def test_retry_inactive():
    """
    If inactive, don't retry.
    """
    num_called = 0

    @stamina.retry(on=Exception)
    async def f():
        nonlocal num_called
        num_called += 1
        raise Exception("passed")

    stamina.set_active(False)

    with pytest.raises(Exception, match="passed"):
        await f()

    assert 1 == num_called


async def test_retry_inactive_ok():
    """
    If inactive, the happy path still works.
    """
    num_called = 0

    @stamina.retry(on=Exception)
    async def f():
        nonlocal num_called
        num_called += 1

    stamina.set_active(False)

    await f()

    assert 1 == num_called


async def test_retry_block():
    """
    Async retry_context blocks are retried.
    """
    num_called = 0

    async for attempt in stamina.retry_context(on=ValueError, wait_max=0):
        with attempt:
            num_called += 1

            assert num_called == attempt.num

            if num_called < 2:
                raise ValueError

    assert 2 == num_called


async def test_retry_blocks_can_be_disabled():
    """
    Async context retries respect the config.
    """
    stamina.set_active(False)
    num_called = 0

    with pytest.raises(Exception, match="passed"):  # noqa: PT012
        async for attempt in stamina.retry_context(on=Exception, attempts=2):
            with attempt:
                num_called += 1
                raise Exception("passed")

    assert 1 == num_called



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from .instrumentation import RetryDetails, RetryHook


__all__ = ["RetryHook", "RetryDetails"]



# SPDX-FileCopyrightText: 2022 Hynek Schlawack <hs@ox.cx>
#
# SPDX-License-Identifier: MIT

import pytest

import stamina


structlog = pytest.importorskip("structlog")


@pytest.fixture(name="log_output")
def _log_output():
    from structlog.testing import LogCapture

    log_output = LogCapture()
    structlog.configure(processors=[log_output])

    return log_output.entries


def test_decorator_sync(log_output):
    """
    Retries decorators log correct name / arguments.
    """

    @stamina.retry(on=ValueError, wait_max=0, attempts=2)
    def f():
        raise ValueError

    with pytest.raises(ValueError):
        f()

    assert [
        {
            "args": (),
            "retry_num": 1,
            "wait_for": 0.0,
            "waited_so_far": 0.0,
            "callable": "tests.test_structlog.test_decorator_sync.<locals>.f",
            "caused_by": "ValueError()",
            "event": "stamina.retry_scheduled",
            "kwargs": {},
            "log_level": "warning",
        },
    ] == log_output


@pytest.mark.anyio()
async def test_decorator_async(log_output):
    """
    Retries decorators log correct name / arguments.
    """

    @stamina.retry(on=ValueError, wait_max=0, attempts=2)
    async def f():
        raise ValueError

    with pytest.raises(ValueError):
        await f()

    assert [
        {
            "args": (),
            "retry_num": 1,
            "wait_for": 0.0,
            "waited_so_far": 0.0,
            "callable": "tests.test_structlog.test_decorator_async.<locals>.f",
            "caused_by": "ValueError()",
            "event": "stamina.retry_scheduled",
            "kwargs": {},
            "log_level": "warning",
        },
    ] == log_output


def test_context_sync(log_output):
    """
    Retries context blocks log correct name / arguments.
    """
    from tests.test_sync import test_retry_block

    test_retry_block()

    assert [
        {
            "callable": "<context block>",
            "retry_num": 1,
            "wait_for": 0.0,
            "waited_so_far": 0.0,
            "caused_by": "ValueError()",
            "args": (),
            "kwargs": {},
            "event": "stamina.retry_scheduled",
            "log_level": "warning",
        }
    ] == log_output


@pytest.mark.anyio()
async def test_context_async(log_output):
    """
    Retries context blocks log correct name / arguments.
    """
    from tests.test_async import test_retry_block

    await test_retry_block()

    assert [
        {
            "callable": "<context block>",
            "retry_num": 1,
            "wait_for": 0.0,
            "waited_so_far": 0.0,
            "caused_by": "ValueError()",
            "args": (),
            "kwargs": {},
            "event": "stamina.retry_scheduled",
            "log_level": "warning",
        }
    ] == log_output
